#
#           Yasuharu Okamoto, 2020.08
#
import tkinter as tk
import tkinter.filedialog as fd
import pandas as pd
import numpy as np
import pystan
import matplotlib.pyplot as plt
import seaborn as sb

def askinfilename(msg = None, types = [('', '*.*')]):
    """
            入力用ファイル名の設定
    """
    rt = tk.Tk()
    rt.withdraw()
    flnm = fd.askopenfilename(title = msg, filetypes = types)
    rt.destroy()
    return flnm

def askoutfilename(msg = None, types = [('', '*.*')], deftype = '.txt'):
    """
            出力用ファイル名の設定
    """
    rt = tk.Tk()
    rt.withdraw()
    flnm = fd.asksaveasfilename(title = msg, filetypes = types,
                                           defaultextension = deftype)
    rt.destroy()
    return flnm

finnm = askinfilename('入力データExcelファイル', types = [('Excel(*.xlsx)', '*.xlsx')])
foutnm = askoutfilename(msg = '画像保存ファイル', types = [('画像 (*.png)', '*.png')],
                       deftype = '.png')
ftxtoutnm = askoutfilename(msg = 'テキスト出力ファイル', types = [('Text (*.txt)', '*.txt')],
                           deftype = '.txt')

fout = open(ftxtoutnm, 'w')
fout.write('入力データファイル = {}\n'.format(finnm))

xlsx_fl = pd.ExcelFile(finnm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values

KC = len(data[0]) - 1
print('KC = ', KC)
KR = len(data) - 1
print('KR = ', KR)

NameC = data[0][1:]
NameR = []
for i in range(KR):
    NameR.append(data[i + 1][0])
print('NameC: ', NameC)
print('NameR: ', NameR)

fout.write('\n行変数：{}\n'.format(NameR))
fout.write('列変数：{}\n'.format(NameC))

X = np.empty((KR, KC), dtype = int)
for i in range(KR):
    for j in range(KC):
        X[i][j] = data[i + 1][j + 1]

for i in range(KR):
    for j in range(KC):
        print('{0:>10d}'.format(X[i][j]), end = '')
    print()

StanData = {'KR': KR, 'KC': KC, 'X': X}

sm = pystan.StanModel(file = 'PoissonModel.stan')
fit = sm.sampling(data = StanData, n_jobs = 1) 
print('\n', fit)

fout.write('\n{}\n'.format(fit))

SavedFiles = []

#
#       主効果と交互作用
#
n_samples = len(fit['z'])
r_e = np.empty((n_samples, KR))
c_e = np.empty((n_samples, KC))
rc_e = np.empty((n_samples, KR, KC))
i = 0
for vz in fit['z']:
    zdd = vz.mean()             #   g..
    zdc = vz.mean(axis = 0)     #   g.j
    zrd = vz.mean(axis = 1)     #   gi.
    r_e[i] = zrd - zdd          #   行主効果
    c_e[i] = zdc - zdd          #   列主効果
    rc_e[i] = vz - zrd.reshape((KR, 1)) - zdc + zdd     #   交互作用
    i += 1

#       日本語表記の設定
plt.rcParams['font.family'] = ['Yu Gothic', 'Noto Sans CJK JP', 'Hiragino Maru Gothic Pro']

plt.title('行主効果', fontsize = 20)
fout.write('\n行主効果\n')
for i in range(KR):
    sb.kdeplot(r_e[:, i], label = NameR[i])
    ra_mean = r_e[:, i].mean()
    ra_025p, ra_25p, ra_50p, ra_75p, ra_975p = \
        np.percentile(r_e[:, i], [2.5, 25, 50, 75, 97.5])
    fout.write('\n    {}\n'.format(NameR[i]))
    fout.write('      平均値 = {0:>.3f}\n'.format(ra_mean))
    fout.write('      中央値 = {0:>.3f}\n'.format(ra_50p))
    fout.write('      Q1 = {0:>.3f},   Q3 = {1:>.3f}\n'.format(ra_25p, ra_75p))
    fout.write('      95%確信区間 = [{0:>.3f}, {1:>.3f}]\n'.format(ra_025p, ra_975p))

plt.yticks([])
plt.legend()
plt.tight_layout()
plt.savefig(foutnm[:-4] + '_ra.png')
SavedFiles.append(foutnm[:-4] + '_ra.png')
plt.show()

plt.title('列主効果', fontsize = 20)
fout.write('\n列主効果\n')
for j in range(KC ):
    sb.kdeplot(c_e[:, j], label = NameC[j])
    ca_mean = c_e[:, j].mean()
    ca_025p, ca_25p, ca_50p, ca_75p, ca_975p = \
        np.percentile(c_e[:, j], [2.5, 25, 50, 75, 97.5])
    fout.write('\n    {}\n'.format(NameC[j]))
    fout.write('      平均値 = {0:>.3f}\n'.format(ca_mean))
    fout.write('      中央値 = {0:>.3f}\n'.format(ca_50p))
    fout.write('      Q1 = {0:>.3f},   Q3 = {1:>.3f}\n'.format(ca_25p, ca_75p))
    fout.write('      95%確信区間 = [{0:>.3f}, {1:>.3f}]\n'.format(ca_025p, ca_975p))
plt.yticks([])
plt.legend()
plt.tight_layout()
plt.savefig(foutnm[:-4] + '_ca.png')
SavedFiles.append(foutnm[:-4] + '_ca.png')
plt.show()

plt.title('交互作用', fontsize = 20)
fout.write('\n交互作用\n')
for i in range(KR):
    for j in range(KC):
        sb.kdeplot(rc_e[:, i, j], label = r'{0}$\times${1}'.format(
            NameR[i], NameC[j]))
        ga_mean = rc_e[:, i, j].mean()
        ga_025p, ga_25p, ga_50p, ga_75p, ga_975p = \
            np.percentile(rc_e[:, i, j], [2.5, 25, 50, 75, 97.5])
        fout.write('\n    {0}×{1}\n'.format(
                    NameR[i], NameC[j]))
        fout.write('      平均値 = {0:>.3f}\n'.format(ga_mean))
        fout.write('      中央値 = {0:>.3f}\n'.format(ga_50p))
        fout.write('      Q1 = {0:>.3f},   Q3 = {1:>.3f}\n'.format(ga_25p, ga_75p))
        fout.write('      95%確信区間 = [{0:>.3f}, {1:>.3f}]\n'.format(ga_025p, ga_975p))
        
plt.yticks([])
plt.legend()
plt.tight_layout()
plt.savefig(foutnm[:-4] + '_rca.png')
SavedFiles.append(foutnm[:-4] + '_rca.png')
plt.show()

fout.write('\n\nλijの中央値:\n')
lmbd_med = np.empty((KR, KC))
for i in range(KR):
    for j in range(KC):
        lmbd_med[i][j] = np.percentile(fit['lmbd'][:, i, j], 50)
        fout.write('\nλ[{0}, {1}] = {2:.3f}\n'.format(NameR[i], NameC[j], lmbd_med[i][j]))

plt.title('$\hat\lambda_{ij}$の中央値', fontsize = 20)
for i in range(KR):
    plt.plot(range(KC), lmbd_med[i], label = NameR[i])
plt.ylabel('$\hat\lambda_{ij}$')
plt.xticks(range(KC), NameC)
plt.legend()
plt.savefig(foutnm[:-4] + '.lr.png')
SavedFiles.append(foutnm[:-4] + '_lr.png')
plt.show()

plt.title('$\hat\lambda_{ij}$の中央値', fontsize = 20)
for j in range(KC):
    plt.plot(range(KR), lmbd_med[:, j], label = NameC[j])
plt.ylabel('$\hat\lambda_{ij}$')
plt.xticks(range(KR), NameR)
plt.legend()
plt.savefig(foutnm[:-4] + '_pc.png')
SavedFiles.append(foutnm[:-4] + '_pc.png')
plt.show()

fout.close()
SavedFiles.append(ftxtoutnm)

print('\n保存ファイル：')
for nm in SavedFiles:
    print('    ', nm)
print()
